# packer.py

from typing import List, Optional, Union
from .item import START_POSITION, Axis
from .bin import Bin
from .item import Item

class Packer:
    def __init__(self):
        """Initialize the packer with empty bins, items, and unfit items lists."""
        self._bins: List[Bin] = []
        self._items: List[Item] = []
        self._unfit_items: List[Item] = []

    @property
    def bins(self) -> List[Bin]:
        """Get list of bins."""
        return self._bins

    @property
    def items(self) -> List[Item]:
        """Get list of items to be packed."""
        return self._items

    @property
    def unfit_items(self) -> List[Item]:
        """Get list of items that couldn't be packed."""
        return self._unfit_items

    def add_bin(self, bin_: Bin) -> None:
        """Add a bin to the packer."""
        self._bins.append(bin_)

    def add_item(self, item: Item) -> None:
        """Add an item to be packed."""
        self._items.append(item)

    def find_fitted_bin(self, item: Item) -> Optional[Bin]:
        """
        Find the first bin that the item fits in.
        
        Args:
            item: Item to find a bin for
            
        Returns:
            Optional[Bin]: Bin that fits the item, or None if no bin fits
        """
        for bin_ in self._bins:
            if not bin_.put_item(item, START_POSITION):
                continue
            
            if len(bin_.items) == 1 and bin_.items[0] == item:
                bin_.items = []
            
            return bin_
        return None

    def get_bigger_bin_than(self, other_bin: Bin) -> Optional[Bin]:
        """
        Find the first bin that's bigger than the given bin.
        
        Args:
            other_bin: Bin to compare against
            
        Returns:
            Optional[Bin]: First bin with larger volume, or None if none found
        """
        return next((b for b in self._bins if b.volume > other_bin.volume), None)

    def unfit_item(self) -> None:
        """Move the first item to the unfit items list."""
        if not self.items:
            return
        self.unfit_items.append(self.items.pop(0))

    def pack_to_bin(self, bin_: Bin, items: List[Item]) -> List[Item]:
        """
        Pack items into the given bin.
        
        Args:
            bin_: Bin to pack items into
            items: Items to pack
            
        Returns:
            List[Item]: Items that couldn't be packed
        """
        b2 = None
        unpacked = []
        
        # Try to fit the first item
        fit = bin_.put_item(items[0], START_POSITION)
        if not fit:
            b2 = self.get_bigger_bin_than(bin_)
            if b2:
                return self.pack_to_bin(b2, items)
            return self.items

        # Pack remaining items
        for item in self.items[1:]:
            fitted = False
            
            # Try different axes for positioning
            for axis in [ Axis.height, Axis.width, Axis.depth, ]:
                if fitted:
                    break
                    
                for item_b in bin_.items:
                    if axis == Axis.width:
                        item_position = [
                            item_b.position[0] + item_b.dimension[0],
                            item_b.position[1],
                            item_b.position[2]
                        ]
                    elif axis == Axis.depth:  # Axis.depth
                        item_position = [
                            item_b.position[0],
                            item_b.position[1],
                            item_b.position[2] + item_b.dimension[2]
                        ]
                    else:
                        item_position = [
                            item_b.position[0],
                            item_b.position[1] + item_b.dimension[1],
                            item_b.position[2]
                        ]

                    if bin_.put_item(item, item_position):
                        fitted = True
                        break

            if not fitted:
                while b2 is not None:
                    b2 = self.get_bigger_bin_than(bin_)
                    if b2:
                        b2.items.append(item)
                        left = self.pack_to_bin(b2, b2.items)
                        if not left:
                            bin_ = b2
                            fitted = True
                            break

                if not fitted:
                    unpacked.append(item)

        return unpacked

    def pack(self) -> None:
        """Pack all items into bins."""
        # Sort bins smallest to largest
        self._bins.sort(key=lambda x: x.volume)
        
        # Sort items largest to smallest
        self._items.sort(key=lambda x: x.volume, reverse=True)
        
        while self._items:
            bin_ = self.find_fitted_bin(self._items[0])
            if not bin_:
                self.unfit_item()
                continue
            
            self._items = self.pack_to_bin(bin_, self._items)


# item.py

from enum import IntEnum
from typing import List, Tuple
from .box import Box

class RotationType(IntEnum):
    whd = 0  # width, height, depth
    hwd = 1  # height, width, depth
    hdw = 2  # height, depth, width
    dhw = 3  # depth, height, width
    dwh = 4  # depth, width, height
    wdh = 5  # width, depth, height

class Axis(IntEnum):
    width = 0
    height = 1
    depth = 2

START_POSITION = [0, 0, 0]

ROTATION_TYPE_STRINGS = {
    RotationType.whd: '(w, h, d)',
    RotationType.hwd: '(h, w, d)',
    RotationType.hdw: '(h, d, w)',
    RotationType.dhw: '(d, h, w)',
    RotationType.dwh: '(d, w, h)',
    RotationType.wdh: '(w, d, h)',
}

def rect_intersect(item1: 'Item', item2: 'Item', x: Axis, y: Axis) -> bool:
    """
    Check if two items intersect in the given plane.
    
    Args:
        item1: First item to check
        item2: Second item to check
        x: First axis to check
        y: Second axis to check
        
    Returns:
        bool: True if items intersect, False otherwise
    """
    d1 = item1.dimension
    d2 = item2.dimension
    
    cx1 = item1.position[x] + d1[x] / 2
    cy1 = item1.position[y] + d1[y] / 2
    cx2 = item2.position[x] + d2[x] / 2
    cy2 = item2.position[y] + d2[y] / 2
    
    ix = max(cx1, cx2) - min(cx1, cx2)
    iy = max(cy1, cy2) - min(cy1, cy2)
    
    return ix < (d1[x] + d2[x]) / 2 and iy < (d1[y] + d2[y]) / 2

class Item(Box):
    def __init__(self, name: str, w: float, h: float, d: float, allowed_rotations: List[RotationType] = None, color: str = '#000000'):
        """
        Initialize an Item with name, dimensions, allowed rotations, and color.
        
        Args:
            name (str): Name of the item
            w (float): Width of the item
            h (float): Height of the item
            d (float): Depth of the item
            allowed_rotations (List[RotationType], optional): List of allowed rotations. Defaults to all rotations.
            color (str, optional): Color of the item. Defaults to '#000000'.
        """
        super().__init__(name, w, h, d)
        self._allowed_rotations = allowed_rotations if allowed_rotations is not None else [
            RotationType.whd,
            RotationType.hwd,
            RotationType.hdw,
            RotationType.dhw,
            RotationType.dwh,
            RotationType.wdh,
        ]
        self._rotation_type = self._allowed_rotations[0]
        self._position: List[float] = []  # x, y, z
        self.color = color # Color of the item

    @property
    def allowed_rotations(self) -> List[RotationType]:
        """Get allowed rotation types for this item."""
        return self._allowed_rotations

    @property
    def rotation_type(self) -> RotationType:
        """Get current rotation type."""
        return self._rotation_type

    @rotation_type.setter
    def rotation_type(self, type_: RotationType):
        """Set current rotation type."""
        self._rotation_type = type_

    @property
    def position(self) -> List[float]:
        """Get current position."""
        return self._position

    @position.setter
    def position(self, position: List[float]):
        """Set current position."""
        self._position = position

    @property
    def rotation_type_string(self) -> str:
        """Get string representation of current rotation type."""
        return ROTATION_TYPE_STRINGS[self._rotation_type]

    @property
    def dimension(self) -> List[float]:
        """Get dimensions based on current rotation type."""
        if self._rotation_type == RotationType.whd:
            return [self.width, self.height, self.depth]
        elif self._rotation_type == RotationType.hwd:
            return [self.height, self.width, self.depth]
        elif self._rotation_type == RotationType.hdw:
            return [self.height, self.depth, self.width]
        elif self._rotation_type == RotationType.dhw:
            return [self.depth, self.height, self.width]
        elif self._rotation_type == RotationType.dwh:
            return [self.depth, self.width, self.height]
        elif self._rotation_type == RotationType.wdh:
            return [self.width, self.depth, self.height]

    def does_intersect(self, other: 'Item') -> bool:
        """
        Check if this item intersects with another item.
        
        Args:
            other: Item to check intersection with
            
        Returns:
            bool: True if items intersect, False otherwise
        """
        return (rect_intersect(self, other, Axis.width, Axis.height) and
                rect_intersect(self, other, Axis.height, Axis.depth) and
                rect_intersect(self, other, Axis.width, Axis.depth))

    def __str__(self) -> str:
        """String representation of the item."""
        return f"Item: {self.name} ({self.rotation_type_string} = {' x '.join(map(str, self.dimension))})"


# utils.py

from typing import Union
import math

# Precision to retain in factored_integer()
FACTOR = 5

def factored_integer(value: Union[int, float]) -> int:
    """
    Factor a number by FACTOR and round to the nearest whole number.
    
    Args:
        value (Union[int, float]): The number to be factored
        
    Returns:
        int: The factored and rounded number
        
    Example:
        >>> factored_integer(1.23456)
        123456
        >>> factored_integer(1.23454)
        123454
    """
    return round(value * (10 ** FACTOR))


import unittest
from src.bin_packer import Packer, Bin, Item

class TestPacker(unittest.TestCase):

    def setUp(self):
        self.test_datas = [
            {
                "name": "Edge case that needs rotation.",
                "bins": [
                    Bin('Le grande box', 100, 100, 300),
                ],
                "items": [
                    Item('Item 1', 150, 50, 50),
                ],
                "expectation": lambda packer: len(packer.bins[0].items) == 1 and len(packer.unfit_items) == 0,
            },
            {
                "name": "Edge case with only rotation 3 and 0 enabled.",
                "bins": [
                    Bin('Le grande box', 100, 100, 300),
                ],
                "items": [
                    Item('Item 1', 150, 50, 50),
                ],
                "expectation": lambda packer: len(packer.bins[0].items) == 1 and len(packer.unfit_items) == 0,
            },
            {
                "name": "Test three items fit into smaller bin after being rotated.",
                "bins": [
                    Bin("1. Le petite box", 296, 296, 8),
                    Bin("2. Le grande box", 2960, 2960, 80),
                ],
                "items": [
                    Item("Item 1", 250, 250, 2),
                    Item("Item 2", 250, 2, 250),
                    Item("Item 3", 2, 250, 250),
                ],
                "expectation": lambda packer: (
                    packer.bins[0].name == '1. Le petite box' and
                    len(packer.bins[0].items) == 3 and
                    len(packer.bins[1].items) == 0 and
                    len(packer.unfit_items) == 0
                ),
            },
            {
                "name": "Test three items fit into larger bin.",
                "bins": [
                    Bin("1. Le petite box", 296, 296, 8),
                    Bin("2. Le grande box", 2960, 2960, 80),
                ],
                "items": [
                    Item("Item 1", 2500, 2500, 20),
                    Item("Item 2", 2500, 2500, 20),
                    Item("Item 3", 2500, 2500, 20),
                ],
                "expectation": lambda packer: (
                    packer.bins[0].name == '1. Le petite box' and
                    len(packer.bins[0].items) == 0 and
                    len(packer.bins[1].items) == 3 and
                    len(packer.unfit_items) == 0
                ),
            },
            {
                "name": "1 bin with 7 items fit into.",
                "bins": [
                    Bin("Bin 1", 220, 160, 100),
                ],
                "items": [
                    Item("Item 1", 20, 100, 30),
                    Item("Item 2", 100, 20, 30),
                    Item("Item 3", 20, 100, 30),
                    Item("Item 4", 100, 20, 30),
                    Item("Item 5", 100, 20, 30),
                    Item("Item 6", 100, 100, 30),
                    Item("Item 7", 100, 100, 30),
                ],
                "expectation": lambda packer: len(packer.bins[0].items) == 7 and len(packer.unfit_items) == 0,
            },
            {
                "name": "Big item is packed first.",
                "bins": [
                    Bin("Bin 1", 100, 100, 100),
                ],
                "items": [
                    Item("Item 1", 50, 100, 100),
                    Item("Item 2", 100, 100, 100),
                    Item("Item 3", 50, 100, 100),
                ],
                "expectation": lambda packer: len(packer.bins[0].items) == 1 and len(packer.unfit_items) == 2,
            },
            {
                "name": "Larger items are tried first.",
                "bins": [
                    Bin("Small Bin", 50, 100, 100),
                    Bin("Bigger Bin", 150, 100, 100),
                    Bin("Small Bin", 50, 100, 100),
                ],
                "items": [
                    Item("Item 1 Small", 50, 100, 100),
                    Item("Item 3 Small", 50, 100, 100),
                    Item("Item 3 Small", 50, 100, 100),
                    Item("Item 2 Big", 100, 100, 100),
                ],
                "expectation": lambda packer: (
                    packer.bins[2].name == 'Bigger Bin' and
                    len(packer.bins[2].items) == 2 and
                    packer.bins[0].name == 'Small Bin' and
                    len(packer.bins[0].items) == 1 and
                    len(packer.unfit_items) == 0
                ),
            },
            {
                "name": "First item fits without rotation but needs to be rotated to fit all items.",
                "bins": [
                    Bin('USPS Medium Flat Rate Box (Top Loading)', 11, 8.5, 5.5),
                ],
                "items": [
                    Item('Item 1', 8.1, 5.2, 2.2),
                    Item('Item 2', 8.1, 5.2, 3.3),
                ],
                "expectation": lambda packer: len(packer.bins[0].items) == 2 and len(packer.unfit_items) == 0,
            },
            {
                "name": "Floating point arithmetic is handled correctly.",
                "bins": [
                    Bin("Bin 1", 12, 12, 5.5),
                ],
                "items": [
                    Item("Item 1", 12, 12, 0.005),
                    Item("Item 2", 12, 12, 0.005),
                ],
                "expectation": lambda packer: len(packer.bins[0].items) == 2 and len(packer.unfit_items) == 0,
            },
        ]

    def test_packer(self):
        for test_data in self.test_datas:
            with self.subTest(test_data["name"]):
                packer = Packer()
                for bin_ in test_data["bins"]:
                    packer.add_bin(bin_)
                for item in test_data["items"]:
                    packer.add_item(item)
                packer.pack()
                self.assertTrue(test_data["expectation"](packer))

if __name__ == "__main__":
    unittest.main()



from typing import List, Tuple
from .log import create_logger
from .box import Box

log = create_logger('3D: ')

class Bin(Box):
    def __init__(self, name: str, w: float, h: float, d: float):
        super().__init__(name, w, h, d)
        self._items = []

    @property
    def items(self) -> List:
        return self._items

    @items.setter
    def items(self, items: List) -> None:
        self._items = items

    def score_rotation(self, item, rotation_type: int) -> float:
        """
        Calculate a score for a given item and rotation type.

        Scores are higher for rotations that closest match item dimensions to Bin dimensions.
        For example, rotating the item so the longest side is aligned with the longest Bin side.

        Example (Bin is 11 x 8.5 x 5.5, Item is 8.1 x 5.2 x 5.2):
         Rotation 0:
           8.1 / 11  = 0.736
           5.2 / 8.5 = 0.612
           5.2 / 5.5 = 0.945
           -----------------
           0.736 ** 2 + 0.612 ** 2 + 0.945 ** 2 = 1.809

         Rotation 1:
           8.1 / 8.5 = 0.953
           5.2 / 11 = 0.473
           5.2 / 5.5 = 0.945
           -----------------
           0.953 ** 2 + 0.473 ** 2 + 0.945 ** 2 = 2.025
        """
        item.rotation_type = rotation_type
        d = item.dimension

        # If the item doesn't fit in the Bin
        if self.width < d[0] or self.height < d[1] or self.depth < d[2]:
            return 0

        # Square the results to increase the impact of high values (e.g. > 0.8)
        width_score = pow(d[0] / self.width, 2)
        height_score = pow(d[1] / self.height, 2)
        depth_score = pow(d[2] / self.depth, 2)

        return width_score + height_score + depth_score

    def get_best_rotation_order(self, item) -> List[int]:
        """
        Calculate the best rotation order for a given Item based on score_rotation().
        Returns rotation types sorted by their score, DESC
        """
        rotation_scores = {}
        
        # Score all rotation types
        for rotation in item.allowed_rotations:
            rotation_scores[rotation] = self.score_rotation(item, rotation)

        # Sort the rotation types by score in descending order
        return sorted(rotation_scores.keys(), 
                     key=lambda x: rotation_scores[x], 
                     reverse=True)

    def put_item(self, item, p: Tuple[float, float, float]) -> bool:
        """
        Attempt to place an item in the bin at the specified position.
        Returns True if the item fits, False otherwise.
        """
        fit = False
        rotations = self.get_best_rotation_order(item)
        item.position = p

        for rotation in rotations:
            item.rotation_type = rotation
            d = item.dimension

            if (self.width < p[0] + d[0] or 
                self.height < p[1] + d[1] or 
                self.depth < p[2] + d[2]):
                fit = False
            else:
                fit = True
                for other_item in self.items:
                    if other_item.does_intersect(item):
                        fit = False
                        break
                
                if fit:
                    self.items.append(item)

            log(f'try to putItem {fit} item {str(item)} box {str(self)}')
            
            if fit:
                break

        return fit

    def __str__(self) -> str:
        return f'Bin: {self.name} (W x H x D = {self.width} x {self.height} x {self.depth})'


# log.py

is_log_enabled = False

def enable_log(enable: bool = True) -> None:
    """Enable or disable logging functionality."""
    global is_log_enabled
    is_log_enabled = enable

def create_logger(namespace: str = '3d-bin-packing'):
    """Create a logger function with a specific namespace."""
    def logger(*args):
        return log(namespace, *args)
    return logger

def log(namespace: str, *args) -> None:
    """Log messages with namespace if logging is enabled."""
    if is_log_enabled:
        print(f"DEBUG: {namespace}", *args)


# box.py

from .utils import factored_integer

class Box:
    def __init__(self, name: str, w: float, h: float, d: float):
        """
        Initialize a Box with name and dimensions.
        
        Args:
            name (str): Name of the box
            w (float): Width of the box
            h (float): Height of the box
            d (float): Depth of the box
        """
        self._name = name
        self._width = factored_integer(w)
        self._height = factored_integer(h)
        self._depth = factored_integer(d)

    @property
    def name(self) -> str:
        """Get the name of the box."""
        return self._name

    @property
    def width(self) -> float:
        """Get the width of the box."""
        return self._width

    @property
    def height(self) -> float:
        """Get the height of the box."""
        return self._height

    @property
    def depth(self) -> float:
        """Get the depth of the box."""
        return self._depth

    @property
    def volume(self) -> float:
        """Calculate and return the volume of the box."""
        return self._width * self._height * self._depth